@@ -10,18 +10,19 @@ from ..log.data_model import Status
 from ..report.data_model import VariableResults, DataGeneratorResults  # noqa: F401
 from ..utils.core import pad_arrays_to_consistent_shapes
 from ..warnings import warn
-from ..xml.utils import get_namespaces_for_xml_doc
-from .data_model import (SedDocument, Model, ModelChange, ModelAttributeChange, AddElementModelChange,  # noqa: F401
-                         ReplaceElementModelChange, RemoveElementModelChange, ComputeModelChange,
-                         Task, Report, Plot2D, Plot3D,
-                         DataGenerator, Variable, MATHEMATICAL_FUNCTIONS, RESERVED_MATHEMATICAL_SYMBOLS)
+from ..xml.utils import eval_xpath
+from .data_model import (SedBase, SedIdGroupMixin, SedDocument,  # noqa: F401
+                         Model, ModelLanguagePattern, ModelChange, ModelAttributeChange, AddElementModelChange,
+                         ReplaceElementModelChange, RemoveElementModelChange, ComputeModelChange, SetValueComputeModelChange,
+                         Task, RepeatedTask, Report, Plot2D, Plot3D,
+                         DataGenerator, Variable, MATHEMATICAL_FUNCTIONS, RESERVED_MATHEMATICAL_SYMBOLS, AGGREGATE_MATH_FUNCTIONS,
+                         Range, UniformRange, VectorRange, FunctionalRange, UniformRangeType)
 from .warnings import InconsistentVariableShapesWarning
 from lxml import etree
 import copy
 import evalidate
 import io
 import libsedml  # noqa: F401
-import math
 import numpy
 import os
 import re
@@ -30,7 +31,9 @@ import tempfile
 
 __all__ = [
     'append_all_nested_children_to_doc',
+    'add_namespaces_to_xml_node',
     'convert_xml_node_to_string',
+    'get_variables_for_task',
     'resolve_model_and_apply_xml_changes',
     'resolve_model',
     'apply_changes_to_xml_model',
@@ -45,6 +48,15 @@ __all__ = [
     'remove_algorithm_parameter_changes',
     'replace_complex_data_generators_with_generators_for_individual_variables',
     'remove_plots',
+    'get_first_last_models_executed_by_task',
+    'get_models_referenced_by_task',
+    'get_models_referenced_by_range',
+    'get_models_referenced_by_model_change',
+    'get_range_len',
+    'resolve_range',
+    'get_namespaces_for_sed_object',
+    'get_xml_node_namespace_tag_target',
+    'is_model_language_encoded_in_xml',
 ]
 
 
@@ -108,26 +120,19 @@ def append_all_nested_children_to_doc(doc):
     doc.data_generators += list(data_generators - set(doc.data_generators))
 
 
-def add_namespaces_to_xml_node(node, namespaces, include_default=False):
+def add_namespaces_to_xml_node(node, namespace_prefixes, namespaces):
     """ Add namespaces to an XML node
 
     Args:
         node (:obj:`libsedml.XMLNode`): XML node
+        namespace_prefixes (:obj:`set` of :obj:`str`): namespace prefixes used in the node which aren't defined in the node
         namespaces (:obj:`libsedml.XMLNamespaces`): namespaces for the parent document
-        include_default (:obj:`bool`, optional): include the default (SED-ML) namespace
     """
-    node_namespaces = node.getNamespaces()
-
     for i_ns in range(namespaces.getNumNamespaces()):
         uri = namespaces.getURI(i_ns)
         prefix = namespaces.getPrefix(i_ns)
 
-        if prefix == '' and not include_default:
-            continue
-
-        i_node_ns_prefix = node_namespaces.getIndexByPrefix(prefix)
-
-        if i_node_ns_prefix == -1:
+        if prefix in namespace_prefixes:
             node.addNamespace(uri, prefix)
 
 
@@ -153,11 +158,32 @@ def get_variables_for_task(doc, task):
     Returns:
         :obj:`list` of :obj:`Variable`: variables that task must record
     """
+    data_generators = set()
+    for output in doc.outputs:
+        if isinstance(output, Report):
+            for data_set in output.data_sets:
+                data_generators.add(data_set.data_generator)
+
+        elif isinstance(output, Plot2D):
+            for curve in output.curves:
+                data_generators.add(curve.x_data_generator)
+                data_generators.add(curve.y_data_generator)
+
+        elif isinstance(output, Plot3D):
+            for surface in output.surfaces:
+                data_generators.add(surface.x_data_generator)
+                data_generators.add(surface.y_data_generator)
+                data_generators.add(surface.z_data_generator)
+
+        else:
+            raise NotImplementedError('Output of type {} is not supported'.format(output.__class__.__name__))
+
     variables = set()
-    for data_gen in doc.data_generators:
+    for data_gen in data_generators:
         for var in data_gen.variables:
             if var.task == task:
                 variables.add(var)
+
     return list(variables)
 
 
@@ -200,9 +226,12 @@ def resolve_model_and_apply_xml_changes(model, sed_doc, working_dir,
     temp_model_source = resolve_model(model, sed_doc, working_dir)
 
     # apply changes to model
-    if apply_xml_model_changes:
+    if apply_xml_model_changes and model.language and is_model_language_encoded_in_xml(model.language):
         # read model from file
-        model_etree = etree.parse(model.source)
+        try:
+            model_etree = etree.parse(model.source)
+        except Exception as exception:
+            raise ValueError('The model could not be parsed because the model is not a valid XML document: {}'.format(str(exception)))
 
         if model.changes:
             # apply changes
@@ -303,7 +332,7 @@ def resolve_model(model, sed_doc, working_dir):
 
 
 def apply_changes_to_xml_model(model, model_etree, sed_doc, working_dir,
-                               variable_values=None,
+                               variable_values=None, range_values=None,
                                validate_unique_xml_targets=True):
     """ Modify an XML-encoded model according to a model change
 
@@ -314,34 +343,35 @@ def apply_changes_to_xml_model(model, model_etree, sed_doc, working_dir,
         working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)
         variable_values (:obj:`dict`, optional): dictionary which contains the value of each variable of each
             compute model change
+        range_values (:obj:`dict`, optional): dictionary which contains the value of each range of each
+            set value compute model change
         validate_unique_xml_targets (:obj:`bool`, optional): whether to validate the XML targets match
             uniue objects
     """
-    # get namespaces
-    namespaces = get_namespaces_for_xml_doc(model_etree)
-
-    # get XPATH evaluator
-    xpath_evaluator = etree.XPathEvaluator(model_etree, namespaces=namespaces)
-
-    # apply changes
-    model_etrees = {model.id: model_etree}
     for change in model.changes:
         if isinstance(change, ModelAttributeChange):
 
             # get object to change
             obj_xpath, sep, attr = change.target.rpartition('/@')
             if sep != '/@':
-                raise ValueError('target {} is not a valid XPATH to an attribute of a model element'.format(change.target))
-            objs = xpath_evaluator(obj_xpath)
+                raise ValueError('target {} is not a valid XPath to an attribute of a model element'.format(change.target))
+            objs = eval_xpath(model_etree, obj_xpath, change.target_namespaces)
             if validate_unique_xml_targets and len(objs) != 1:
                 raise ValueError('xpath {} must match a single object'.format(obj_xpath))
 
+            ns_prefix, _, attr = attr.rpartition(':')
+            if ns_prefix:
+                ns = change.target_namespaces.get(ns_prefix, None)
+                if ns is None:
+                    raise ValueError('No namespace is defined with prefix `{}`'.format(ns_prefix))
+                attr = '{{{}}}{}'.format(ns, attr)
+
             # change value
             for obj in objs:
                 obj.set(attr, change.new_value)
 
         elif isinstance(change, AddElementModelChange):
-            parents = xpath_evaluator(change.target)
+            parents = eval_xpath(model_etree, change.target, change.target_namespaces)
 
             if validate_unique_xml_targets and len(parents) != 1:
                 raise ValueError('xpath {} must match a single object'.format(change.target))
@@ -349,14 +379,14 @@ def apply_changes_to_xml_model(model, model_etree, sed_doc, working_dir,
             try:
                 new_elements = etree.fromstring('<root>' + change.new_elements + '</root>').getchildren()
             except etree.XMLSyntaxError as exception:
-                raise ValueError('`{}` is not valid XML. {}'.format(change.new_elements, str(exception)))
+                raise ValueError('`{}` is invalid XML. {}'.format(change.new_elements, str(exception)))
 
             for parent in parents:
                 for new_element in copy.deepcopy(new_elements):
                     parent.append(new_element)
 
         elif isinstance(change, ReplaceElementModelChange):
-            old_elements = xpath_evaluator(change.target)
+            old_elements = eval_xpath(model_etree, change.target, change.target_namespaces)
 
             if validate_unique_xml_targets and len(old_elements) != 1:
                 raise ValueError('xpath {} must match a single object'.format(change.target))
@@ -364,7 +394,7 @@ def apply_changes_to_xml_model(model, model_etree, sed_doc, working_dir,
             try:
                 new_elements = etree.parse(io.StringIO('<root>' + change.new_elements + '</root>')).getroot().getchildren()
             except etree.XMLSyntaxError as exception:
-                raise ValueError('`{}` is not valid XML. {}'.format(change.new_elements, str(exception)))
+                raise ValueError('`{}` is invalid XML. {}'.format(change.new_elements, str(exception)))
 
             for old_element in old_elements:
                 parent = old_element.getparent()
@@ -375,7 +405,7 @@ def apply_changes_to_xml_model(model, model_etree, sed_doc, working_dir,
                     parent.append(new_element)
 
         elif isinstance(change, RemoveElementModelChange):
-            elements = xpath_evaluator(change.target)
+            elements = eval_xpath(model_etree, change.target, change.target_namespaces)
 
             if validate_unique_xml_targets and len(elements) != 1:
                 raise ValueError('xpath {} must match a single object'.format(change.target))
@@ -387,22 +417,36 @@ def apply_changes_to_xml_model(model, model_etree, sed_doc, working_dir,
         elif isinstance(change, ComputeModelChange):
             # get the values of model variables referenced by compute model changes
             if variable_values is None:
-                variable_values = get_values_of_variable_model_xml_targets_of_model_change(change, sed_doc, model_etrees, working_dir)
+                model_etrees = {model.id: model_etree}
+                iter_variable_values = get_values_of_variable_model_xml_targets_of_model_change(change, sed_doc, model_etrees, working_dir)
+            else:
+                iter_variable_values = variable_values
 
             # calculate new value
-            new_value = calc_compute_model_change_new_value(change, variable_values)
+            new_value = calc_compute_model_change_new_value(change, variable_values=iter_variable_values, range_values=range_values)
+            if new_value == int(new_value):
+                new_value = str(int(new_value))
+            else:
+                new_value = str(new_value)
 
             # get object to change
             obj_xpath, sep, attr = change.target.rpartition('/@')
             if sep != '/@':
-                raise ValueError('target {} is not a valid XPATH to an attribute of a model element'.format(change.target))
-            objs = xpath_evaluator(obj_xpath)
+                raise ValueError('target {} is not a valid XPath to an attribute of a model element'.format(change.target))
+            objs = eval_xpath(model_etree, obj_xpath, change.target_namespaces)
             if validate_unique_xml_targets and len(objs) != 1:
                 raise ValueError('xpath {} must match a single object'.format(obj_xpath))
 
+            ns_prefix, _, attr = attr.rpartition(':')
+            if ns_prefix:
+                ns = change.target_namespaces.get(ns_prefix, None)
+                if ns is None:
+                    raise ValueError('No namespace is defined with prefix `{}`'.format(ns_prefix))
+                attr = '{{{}}}{}'.format(ns, attr)
+
             # change value
             for obj in objs:
-                obj.set(attr, str(new_value))
+                obj.set(attr, new_value)
 
         else:
             raise NotImplementedError('Change{} of type {} is not supported'.format(
@@ -456,16 +500,22 @@ def get_value_of_variable_model_xml_targets(variable, model_etrees):
     if not variable.target:
         raise NotImplementedError('Compute model change variable `{}` must have a target'.format(variable.id))
 
+    if variable.target.startswith('#'):
+        raise NotImplementedError('Variable references to data generators are not supported.')
+
     obj_xpath, sep, attr = variable.target.rpartition('/@')
     if sep != '/@':
-        raise ValueError('target {} is not a valid XPATH to an attribute of a model element'.format(variable.target))
+        raise ValueError('target {} is not a valid XPath to an attribute of a model element'.format(variable.target))
 
     et = model_etrees[variable.model.id]
-    namespaces = get_namespaces_for_xml_doc(et)
-    obj = et.xpath(obj_xpath, namespaces=namespaces)
+    obj = eval_xpath(et, obj_xpath, variable.target_namespaces)
     if len(obj) != 1:
         raise ValueError('xpath {} must match a single object in model {}'.format(obj_xpath, variable.model.id))
 
+    ns, _, attr = attr.rpartition(':')
+    if ns:
+        attr = '{{{}}}{}'.format(variable.target_namespaces[ns], attr)
+
     value = obj[0].get(attr)
     if value is None:
         raise ValueError('Target `{}` is not defined in model `{}`.'.format(variable.target, variable.model.id))
@@ -477,13 +527,15 @@ def get_value_of_variable_model_xml_targets(variable, model_etrees):
     return value
 
 
-def calc_compute_model_change_new_value(change, variable_values):
+def calc_compute_model_change_new_value(change, variable_values=None, range_values=None):
     """ Calculate the new value of a compute model change
 
     Args:
         change (:obj:`ComputeModelChange`): change
         variable_values (:obj:`dict`, optional): dictionary which contains the value of each variable of each
             compute model change
+        range_values (:obj:`dict`, optional): dictionary which contains the value of each range of each
+            set value compute model change
 
     Returns:
         :obj:`float`: new value
@@ -491,8 +543,16 @@ def calc_compute_model_change_new_value(change, variable_values):
     compiled_math = compile_math(change.math)
 
     workspace = {}
+
+    if isinstance(change, SetValueComputeModelChange):
+        if change.range:
+            workspace[change.range.id] = range_values.get(change.range.id, None)
+            if workspace[change.range.id] is None:
+                raise ValueError('Value of range `{}` is not defined.'.format(change.range.id))
+
     for param in change.parameters:
         workspace[param.id] = param.value
+
     for var in change.variables:
         workspace[var.id] = variable_values.get(var.id, None)
         if workspace[var.id] is None:
@@ -535,10 +595,15 @@ def calc_data_generator_results(data_generator, variable_results):
         workspace[param.id] = param.value
 
     if not var_shapes:
-        value = eval_math(math, compiled_math, workspace)
+        value = eval_math(data_generator.math, compiled_math, workspace)
         result = numpy.array(value)
 
     else:
+        for aggregate_func in AGGREGATE_MATH_FUNCTIONS:
+            if re.search(aggregate_func + r' *\(', data_generator.math):
+                msg = 'Evaluation of aggregate mathematical functions such as `{}` is not supported.'.format(aggregate_func)
+                raise NotImplementedError(msg)
+
         padded_var_shapes = []
         for var in data_generator.variables:
             var_res = variable_results[var.id]
@@ -575,7 +640,7 @@ def calc_data_generator_results(data_generator, variable_results):
             if not vars_available:
                 continue
 
-            result_el = eval_math(math, compiled_math, workspace)
+            result_el = eval_math(data_generator.math, compiled_math, workspace)
 
             if n_dims == 0:
                 result = numpy.array(result_el)
@@ -668,10 +733,17 @@ def compile_math(math):
     Returns:
         :obj:`_ast.Expression`: compiled expression
     """
+    if isinstance(math, str):
+        math = (
+            math
+            .replace('&&', 'and')
+            .replace('||', 'or')
+        )
+
     math_node = evalidate.evalidate(math,
                                     addnodes=[
                                         'Eq', 'NotEq', 'Gt', 'Lt', 'GtE', 'LtE',
-                                        'Sub', 'Mult', 'Div' 'Pow',
+                                        'Sub', 'Mult', 'Div', 'Pow',
                                         'And', 'Or', 'Not',
                                         'BitAnd', 'BitOr', 'BitXor',
                                         'Call',
@@ -703,8 +775,8 @@ def eval_math(math, compiled_math, workspace):
     try:
         return eval(compiled_math, MATHEMATICAL_FUNCTIONS, dict(**RESERVED_MATHEMATICAL_SYMBOLS, **workspace))
     except Exception as exception:
-        raise ValueError('Expression `{}` could not be evaluated:\n  {}'.format(
-            math, str(exception)))
+        raise ValueError('Expression `{}` could not be evaluated:\n\n  {}\n\n  workspace:\n    {}'.format(
+            math, str(exception), '\n    '.join('{}: {}'.format(key, value) for key, value in workspace.items())))
 
 
 def remove_model_changes(sed_doc):
@@ -790,3 +862,308 @@ def remove_plots(sed_doc):
     for output in list(sed_doc.outputs):
         if isinstance(output, (Plot2D, Plot3D)):
             sed_doc.outputs.remove(output)
+
+
+def get_first_last_models_executed_by_task(task):
+    """ Get the models executed by a task in the order they are executed
+
+    Args:
+        task (:obj:`Task`): task
+
+    Returns:
+        :obj:`tuple` of :obj:`Model`: models
+    """
+    if isinstance(task, Task):
+        return (task.model, task.model)
+
+    elif isinstance(task, RepeatedTask):
+        sub_tasks = sorted(task.sub_tasks, key=lambda sub_task: sub_task.order)
+        return (
+            get_first_last_models_executed_by_task(sub_tasks[0].task)[0],
+            get_first_last_models_executed_by_task(sub_tasks[-1].task)[-1],
+        )
+
+    else:
+        raise NotImplementedError('Tasks of type `{}` are not supported.'.format(task.__class__.__name__))
+
+
+def get_models_referenced_by_task(task):
+    """ Get the models referenced from a task
+
+    Args:
+        task (:obj:`Task`): task
+
+    Returns:
+        :obj:`set` of :obj:`Model`: models
+    """
+    if isinstance(task, Task):
+        return set([task.model])
+
+    elif isinstance(task, RepeatedTask):
+        models = set()
+
+        models.update(get_models_referenced_by_range(task.range))
+
+        for change in task.changes:
+            models.update(get_models_referenced_by_model_change(change))
+
+        for sub_task in task.sub_tasks:
+            models.update(get_models_referenced_by_task(sub_task.task))
+
+        if task.range:
+            models.update(get_models_referenced_by_range(task.range))
+        for range in task.ranges:
+            models.update(get_models_referenced_by_range(range))
+
+        return models
+
+    else:
+        raise NotImplementedError('Tasks of type `{}` are not supported.'.format(task.__class__.__name__))
+
+
+def get_models_referenced_by_model_change(change):
+    """ Get the models referenced from a model change
+
+    Args:
+        change (:obj:`ModelChange`): model change
+
+    Returns:
+        :obj:`set` of :obj:`Model`: models
+    """
+    models = set()
+
+    if isinstance(change, SetValueComputeModelChange):
+        if change.model:
+            models.add(change.model)
+
+        if change.range:
+            models.update(get_models_referenced_by_range(change.range))
+
+        for variable in change.variables:
+            if variable.model:
+                models.add(variable.model)
+
+    elif isinstance(change, ComputeModelChange):
+        for variable in change.variables:
+            if variable.model:
+                models.add(variable.model)
+
+    return models
+
+
+def get_models_referenced_by_range(range):
+    """ Get the models referenced by a range
+
+    Args:
+        range (:obj:`Range`): range
+
+    Returns:
+        :obj:`set` of :obj:`Model`: models
+    """
+    models = set()
+
+    if isinstance(range, FunctionalRange):
+        for variable in range.variables:
+            if variable.model:
+                models.add(variable.model)
+
+        models.update(get_models_referenced_by_range(range.range))
+
+    return models
+
+
+def get_range_len(range):
+    """ Get the length of a range
+
+    Args:
+        range (:obj:`Range`): range
+
+    Returns:
+        :obj:`int`: length of the range
+
+    Raises:
+        :obj:`NotImplementedError`: if range isn't an instance of :obj:`UniformRange`, :obj:`VectorRange`,
+            or :obj:`FunctionalRange`.
+    """
+    if isinstance(range, UniformRange):
+        return range.number_of_steps + 1
+
+    elif isinstance(range, VectorRange):
+        return len(range.values)
+
+    elif isinstance(range, FunctionalRange):
+        return get_range_len(range.range)
+
+    else:
+        raise NotImplementedError('Ranges of type `{}` are not supported.'.format(range.__class__.__name__))
+
+
+def resolve_range(range, model_etrees=None):
+    """ Resolve the values of a range
+
+    Args:
+        range (:obj:`Range`): range
+        model_etrees (:obj:`dict` of :obj:`str` to :obj:`etree._Element`): map from the ids of models to element
+            trees of their sources; required to resolve variables of functional ranges
+
+    Returns:
+        :obj:`list` of :obj:`float`: values of the range
+
+    Raises:
+        :obj:`NotImplementedError`: if range isn't an instance of :obj:`UniformRange`, :obj:`VectorRange`,
+            or :obj:`FunctionalRange`.
+    """
+    if isinstance(range, UniformRange):
+        if range.type == UniformRangeType.linear:
+            return numpy.linspace(range.start, range.end, range.number_of_steps + 1).tolist()
+
+        elif range.type == UniformRangeType.log:
+            return numpy.logspace(numpy.log10(range.start), numpy.log10(range.end), range.number_of_steps + 1).tolist()
+
+        else:
+            raise NotImplementedError('UniformRanges of type `{}` are not supported.'.format(range.type.value))
+
+    elif isinstance(range, VectorRange):
+        return range.values
+
+    elif isinstance(range, FunctionalRange):
+        # compile math
+        compiled_math = compile_math(range.math)
+
+        # setup workspace to evaluate math
+        workspace = {}
+
+        workspace[range.range.id] = None
+
+        for param in range.parameters:
+            workspace[param.id] = param.value
+
+        for var in range.variables:
+            if var.symbol:
+                raise NotImplementedError('Symbols are not supported for variables of functional ranges')
+            if model_etrees[var.model.id] is None:
+                raise NotImplementedError('Functional ranges that involve variables of non-XML-encoded models are not supported.')
+            workspace[var.id] = get_value_of_variable_model_xml_targets(var, model_etrees)
+
+        # calculate the values of the range
+        values = []
+        for child_range_value in resolve_range(range.range, model_etrees=model_etrees):
+            workspace[range.range.id] = child_range_value
+
+            value = eval_math(range.math, compiled_math, workspace)
+            values.append(value)
+
+        # return values
+        return values
+
+    else:
+        raise NotImplementedError('Ranges of type `{}` are not supported.'.format(range.__class__.__name__))
+
+
+def get_namespaces_for_sed_object(obj):
+    """ Get the namespace prefixes and URIs defined for a SED object
+
+    Args:
+        obj (:obj:`libsedml.SedBase`): SED object
+
+    Returns:
+        :obj:`dict`: dictionary that maps the prefixes of namespaces to their URIs
+    """
+    namespaces_obj = obj.getNamespaces()
+    namespaces_dict = {}
+    for i_namespace in range(namespaces_obj.getNumNamespaces()):
+        prefix = namespaces_obj.getPrefix(i_namespace) or None
+        uri = namespaces_obj.getURI(i_namespace)
+        namespaces_dict[prefix] = uri
+    return namespaces_dict
+
+
+def get_xml_node_namespace_tag_target(etree, target_namespaces=None):
+    """ Get the namespace, tag, and target of an XML node
+
+    Args:
+        etree (:obj:`etree._Element`): element tree
+        target_namespaces (:obj:`dict`, optional): dictionary that maps the prefixes of namespaces to their URIs
+
+    Returns:
+        :obj:`tuple`:
+
+            * :obj:`str`: namespace URI
+            * :obj:`str`: namespace prefix
+            * :obj:`str`: tag
+            * :obj:`str`: target for use with SED target XPaths
+            * :obj:`dict`: dictionary that maps the prefixes of namespaces to their URIs
+    """
+    target_namespaces = dict(target_namespaces or {})
+    uri, _, tag = etree.tag.rpartition('}')
+    if uri:
+        uri = uri[1:]
+    if uri:
+        nsmap_rev = {uri: prefix for prefix, uri in etree.nsmap.items()}
+        prefix = nsmap_rev[uri]
+        if not prefix:
+            target_namespaces_rev = {uri: prefix for prefix, uri in target_namespaces.items()}
+            prefix = target_namespaces_rev.get(uri, tag)
+
+        if prefix in target_namespaces:
+            if target_namespaces.get(prefix, None) != uri:
+                prefix += '2'
+                target_namespaces[prefix] = uri
+        else:
+            target_namespaces[prefix] = uri
+
+        target = prefix + ':' + tag
+    else:
+        uri = None
+        prefix = None
+        target = tag
+
+    return (uri, prefix, tag, target, target_namespaces)
+
+
+def is_model_language_encoded_in_xml(language):
+    """ Determine if the model language is encoded in XML
+
+    Args:
+        language (:obj:`str`): language
+
+    Returns:
+        :obj:`bool`: :obj:`True`, if the model language is encoded in XML
+    """
+    return (
+        re.match(ModelLanguagePattern.CellML, language)
+        or re.match(ModelLanguagePattern.NeuroML, language)
+        or re.match(ModelLanguagePattern.SBML, language)
+        or re.match(ModelLanguagePattern.VCML, language)
+    )
+
+
+def get_all_sed_objects(doc, type=(SedBase, SedIdGroupMixin)):
+    """ Get all of the identified objects (instances of :obj:`SedIdGroupMixin`) that
+    belong to a SED document
+
+    Args:
+        doc (:obj:`SedBase`) SED document
+        type (:obj:`type` or :obj:`tuple` of :obj:`type`, optional): type of SED object
+            to return
+    """
+    # get all linked objects
+    objs_to_see = [doc]
+    seen_objs = [doc]
+    while objs_to_see:
+        obj = objs_to_see.pop()
+        for attr in obj.__dict__.values():
+            if isinstance(attr, (SedBase, SedIdGroupMixin)):
+                if attr not in seen_objs:
+                    objs_to_see.append(attr)
+                    seen_objs.append(attr)
+
+            elif isinstance(attr, list):
+                for val in attr:
+                    if isinstance(val, (SedBase, SedIdGroupMixin)):
+                        if val not in seen_objs:
+                            objs_to_see.append(val)
+                            seen_objs.append(val)
+
+    # filter out elements of a specific type
+    return list(filter(lambda obj: isinstance(obj, type), seen_objs))
