@@ -18,6 +18,6 @@ ENVIRONMENT_VARIABLES = {
         description='Policy for substituting alternative algorithms.',
         options=[policy.value for policy, _ in sorted(ALGORITHM_SUBSTITUTION_POLICY_LEVELS.items(), key=lambda policy_level: policy_level[1])],
         default=AlgorithmSubstitutionPolicy.SIMILAR_VARIABLES,
-        more_info_url='https://biosimulators.org/standards/simulator-interfaces',
+        more_info_url='https://biosimulators.org/conventions/simulator-interfaces',
     )
 }
